/* Custom thread-specific data maintainance.  Currently only supports the case
 * when thread ids are an interval of integers except for the first value, and
 * tid1 < tid2 iff thread1 was spawned before thread2 (except perhaps for the
 * first value).
 */

#include <stdio.h>
#include "mpii.h"

#ifdef MPII_TSD

/* Stores the tid of the first thread.  It is handled specially because it
 * usually doesn't relate to other tids.
 */
static int first_tid = -1;

/* Stores the tid of the second thread.  Hence, an index is computed by
 * tid-second_tid+1, unless tid == first_tid.
 */
static int second_tid = -1;

#define map(tid) ((tid) == first_tid ? 0 : (tid)-second_tid+1)

/* Stores the id of the last thread that registered. */
static int last_tid = -1;

/* The total number of threads that will register. */
static int nthread = -1;

/* Called on the first thread before any threads have been spawned.
 * n is the number of threads, including the master.
 */
PRIVATE void MPII_Tsd_master_init (Thread id, int n)
{
    first_tid = last_tid = id;
    nthread = n;
}

/* Called by each spawned thread.  These must be done in order. */
PRIVATE void MPII_Tsd_slave_init (Thread id)
{
    if (first_tid == last_tid)
        second_tid = last_tid = id;
    else if (id == last_tid + 1)
    {
        last_tid++;
        if (last_tid - second_tid + 2 > nthread)
        {
            fprintf (stderr, "MPII_Tsd: Extra threads were spawned!");
            exit (1);
        }
    }
    else
    {
        fprintf (stderr, "MPII_Tsd: Your thread system doesn't obey the expected property that thread ids\n");
        fprintf (stderr, "MPII_Tsd: form an interval of integers, and tid1 < tid2 iff thread1 was spawned\n");
        fprintf (stderr, "MPII_Tsd: before thread2, except perhaps for the very first thread id.  You can\n");
        fprintf (stderr, "MPII_Tsd: try commenting out MPII_TSD in include/threading.h and recompiling\n");
        fprintf (stderr, "MPII_Tsd: TOMPI (make clean; make).  This will use the thread-specific data\n");
        fprintf (stderr, "MPII_Tsd: facility in the thread system.  However, this alternative was likely\n");
        fprintf (stderr, "MPII_Tsd: chosen to avoid bugs in that facility.\n");
        fprintf (stderr, "MPII_Tsd: TIDs: first = %d, second = %d, last = %d, next = %d\n", first_tid, second_tid, last_tid, id);
        exit (1);
    }
}

PRIVATE void MPII_New_tsd (Key *key)
{
    int i;

    *key = (Key) malloc (sizeof (void *) * nthread);
    if (*key == NULL)
    {
        perror ("MPII_Tsd: Couldn't create thread-specific data");
        exit (1);
    }
    
    for (i = 0; i < nthread; i++)
        (*key)[i] = NULL;
}

PRIVATE void *MPII_Get_tsd (Key *key)
{
    int id = thread_id ();
/*
printf ("getting %p[%d->%d]...\n", *key, id, map(id));
printf ("got %p[%d->%d] as %p\n", *key, id, map(id), (*key)[map(id)]);
*/
    return (*key)[map(id)];
}

PRIVATE void MPII_Set_tsd (Key *key, void *val)
{
    int id = thread_id ();
/*
printf ("setting %p[%d->%d] to %p\n", *key, id, map(id), val);
*/
    (*key)[map(id)] = val;
}

#endif
