/* Efficient queue package.  Allocates in chunks and uses free list.  Grows
 * dynamically and never shrinks.
 *
 * Routines return 1 if there was an allocation error or 0 on success.
 */

/* To take outside of this MPI thing:
 *    - remove the following include line
 *    - replace MPII_Msg_queue with general "Queue" type (see mpii.h)
 *    - replace MPII_Msg with desired Data type
 *    - remove MPII_ prefixes from routines (other than grow())
 */
#include "mpii.h"

#define BLOCK_SIZE 64

#define type2str(type) ((type) == MSG_AVAIL ? "msg-avail" : "took-msg")

PRIVATE int MPII_queue_init (MPII_Msg_queue *qu)
{
   int i;

   qu->max = BLOCK_SIZE;
   qu->head = qu->tail = qu->dirty = qu->dirty_prev = -1;

   qu->q = (MPII_Msg *) malloc (sizeof (MPII_Msg) * BLOCK_SIZE);
   if (qu->q == NULL)
      return 1;
   qu->freelist = (int *) malloc (sizeof (int) * BLOCK_SIZE);
   if (qu->freelist == NULL)
      return 1;
   qu->next = (int *) malloc (sizeof (int) * BLOCK_SIZE);
   if (qu->next == NULL)
      return 1;

   qu->nfree = BLOCK_SIZE;
   for (i = 0; i < BLOCK_SIZE; i++)
      qu->freelist[i] = i;

   return 0;
}

static int grow (MPII_Msg_queue *qu)
{
   MPII_Msg *q;
   int *freelist, *next;
   int i, j, newsize;

   newsize = qu->max + BLOCK_SIZE;
   q = (MPII_Msg *) malloc (sizeof (MPII_Msg) * newsize);
   if (q == NULL)
      return 1;
   freelist = (int *) malloc (sizeof (int) * newsize);
   if (freelist == NULL)
      return 1;
   next = (int *) malloc (sizeof (int) * newsize);
   if (next == NULL)
      return 1;

   for (i = 0; i < qu->max; i++)
   {
      q[i] = qu->q[i];
      next[i] = qu->next[i];
   }
   for (j = 0, i = qu->max; i < newsize; i++, j++)
      freelist[j] = i;
   for (i = 0; i < qu->nfree; i++, j++)
      freelist[j] = qu->freelist[i];

   free (qu->q);
   free (qu->freelist);
   free (qu->next);

   qu->q = q;
   qu->freelist = freelist;
   qu->next = next;
   qu->max = newsize;
   qu->nfree = j;
   /* Don't modify head or tail */

   return 0;
}

PRIVATE int MPII_enqueue (MPII_Msg_queue *qu, MPII_Msg *data)
{
   int pos;

   if (qu->nfree == 0)
      if (grow (qu))
         return 1;

   pos = qu->freelist[--qu->nfree];
   qu->q[pos] = *data;
   qu->next[pos] = -1;

#  if DEBUG_ENQUEUE >= 1
      printf ("enqueue: Adding %s.\n", type2str (qu->q[pos].type));
#  endif

   if (qu->head < 0)
   {
#     if DEBUG_ENQUEUE >= 2
         printf ("enqueue: That was the only element in the list (%d).\n", pos);
#     endif
      qu->head = qu->tail = pos;

      if (qu->dirty < 0)
      {
         qu->dirty = pos;
         qu->dirty_prev = -1;
      }
   }
   else
   {
#     if DEBUG_ENQUEUE >= 2
         printf ("enqueue: Inserted %d after %d (%s)\n", pos, qu->tail,
               type2str (qu->q[qu->tail].type));
#     endif
      qu->next[qu->tail] = pos;

      if (qu->dirty < 0)
      {
         qu->dirty = pos;
         qu->dirty_prev = qu->tail;
      }

      qu->tail = pos;
   }
}

/* Return value is different here: returns 1 if a match was found, 0 otherwise
 */
PRIVATE int MPII_queue_search (int *retry, MPII_Msg_queue *qu, int (*match) (void *, MPII_Msg *), void *arg, MPII_Msg *result)
{
   int pos, last;

   if (! (*retry))
   {
#     if DEBUG_QUEUE_SEARCH >= 3
         printf ("queue_search: Retry mode off, so starting at head %d\n",
               qu->head);
#     endif
      *retry = 1;
      qu->dirty = qu->head;
      qu->dirty_prev = -1;
   }
#  if DEBUG_QUEUE_SEARCH >= 3
      else
         printf ("queue_search: Retry mode on, so starting at dirty %d\n",
               qu->dirty);
#  endif

   for (last = qu->dirty_prev, pos = qu->dirty; pos >= 0; pos = qu->next[pos])
   {
#     if DEBUG_QUEUE_SEARCH >= 2
         printf ("queue_search: Searching through %d (%s, next: %d)\n",
               pos, type2str (qu->q[pos].type), qu->next[pos]);
#     endif
      if (match (arg, &(qu->q[pos])))
      {
#        if DEBUG_QUEUE_SEARCH >= 1
            printf ("queue_search: Matched %s.\n", type2str (qu->q[pos].type));
#        endif
         if (last < 0)
            qu->head = qu->next[pos];
         else
            qu->next[last] = qu->next[pos];
         if (pos == qu->tail)
            qu->tail = last;
         qu->freelist[qu->nfree++] = pos;
         qu->dirty = qu->next[pos];
         qu->dirty_prev = last;
         *result = qu->q[pos];
         return 1;
      }
      last = pos;
   }

   qu->dirty = qu->dirty_prev = -1;
   return 0;
}

